/*  main.cpp
	mp3player on sc12

	31.03.2002: ab, BIOS Wrapper 0xA0/0x28 findfirst/findnext

	Copyright (c)2002    (ab)Andreas Brachold, a.brachold@berlin.de

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License as
	published by the Free Software Foundation; either version 2 of
	the License, or (at your option) any later version. Read the
	full License at http://www.gnu.org/copyleft for more details.
*/

#include <DOS.H>
#include "fast_findnext.h"

// Workaround for filenames shorter then 8(12) character
// on bios V1.02b is null-terminator for missing
static void trim(char *str, int length)
{
	for (int i = length - 1; i>=0; i--)
		if (str[i] == ' ' || str[i] == '\0')
			str[i] = '\0';
}

int fast_findfirst(const char *pathname, struct filefind *ff, int attrib)
{
   union REGS inregs, outregs;
   struct SREGS sregs;

   inregs.h.ah = 0x28;
   inregs.x.cx = attrib;
   inregs.x.bx = FP_SEG(pathname);
   inregs.x.si = FP_OFF(pathname);
   sregs.es    = FP_SEG(ff);
   inregs.x.di = FP_OFF(ff);
   int86x( 0xA0, &inregs, &outregs, &sregs );
   if (outregs.x.dx || outregs.h.al!=1) return -1;
   else
   trim(ff->filename,sizeof(ff->filename));
   return 0;
}


int fast_findnext(struct filefind *ff)
{
   union REGS inregs, outregs;
   struct SREGS sregs;

   inregs.h.ah = 0x29;
   sregs.es    = FP_SEG(ff);
   inregs.x.di = FP_OFF(ff);
   int86x( 0xA0, &inregs, &outregs, &sregs );
   if (outregs.x.dx || outregs.h.al!=1) return -1;
   else
   trim(ff->filename,sizeof(ff->filename));
   return 0;
}


int fast_finddone(struct filefind *ff)
{
   union REGS inregs, outregs;
   struct SREGS sregs;

   inregs.h.ah = 0x30;
   sregs.es    = FP_SEG(ff);
   inregs.x.di = FP_OFF(ff);
   int86x( 0xA0, &inregs, &outregs, &sregs );
   if (outregs.x.dx || outregs.h.al!=1) return -1;
   else return 0;
}

#ifdef TEST
int main(int paramcnt, char *param[])
{
   struct filefind ff;

   int done;
   char *file;

   if (paramcnt>1)
   {
     file=param[1];
   }
   else file="*.*";

   printf("Directory listing of %s (Using SC12 BIOS Ints FAST_FF!)\n", file);
   done = fast_findfirst(file,&ff,0);
   while (!done)
   {
      printf("  %s.%s\n", ff.filename, ff.fileext);
      done = fast_findnext(&ff);
   }
   fast_finddone(&ff);

   return 0;
}
#endif

