/*  fastlist.cpp
	mp3player playlist support

	15.05.2001: tk, initial implementation.
	03.01.2002: ab, rewritten for recursive searching withhin directory.
	04.01.2002: ab, splitt backend and frontend (now main.cpp).
	12.02.2002: ab, add random seek in playlist
	22.02.2002: ab, change Keys from RC_ENUM to VK
	23.02.2002: ab, remove some static variables
	31.03.2002: ab, use pure BIOS findfirst/findnext
	01.04.2002: ab, use given filesize from findfirst/findnext struct
	04.04.2002: ab, remove useless memset for first displayline
	23.04.2002: ab, fix STOP -> jump now to start of file on keypress
	23.04.2002: ab, implement skip to next directory
	02.05.2002: ab, fix bug in skip to prev directory

	Copyright (c)2000 by Thomas Kindler, thomas.kindler@gmx.de
	2002    (ab)Andreas Brachold, a.brachold@berlin.de

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License as
	published by the Free Software Foundation; either version 2 of
	the License, or (at your option) any later version. Read the
	full License at http://www.gnu.org/copyleft for more details.
*/

// include files -----
//
#include <stdio.h>
#include <dir.h>
#include <stdlib.h>
#include <alloc.h>

#include "sc12/rtos.h"
#include "id3v1.h"
#include "vs1001.h"
#include "mpbuf.h"

#include "main.h"
#include "playlist.h"
#include "fast_findnext.h"

/// The one and only Playlist for all founded files
struct _playlist
{
	/// store a duplicated pointer with the filename
  	char *filename;
	/// played retry for this file, (need for fair random)
	unsigned char nPlayed;
	/// number of directory of the this file ( need for retry directory)
	unsigned char nDirectory;
	/// remember given filesize from fastfilefind (we need no more do ftellg)
	unsigned long nFilesize;
} static playlist[500] = {0};
// 500 is limit of total founded files (2GByte FAT16 LIMIT / 4 MByte on a file)

/** Initialize list with empty filenames
*/
void fastlist_init()
{
		for(unsigned int i= 0;i<memberof(playlist);++i)
          playlist[i].filename = NULL;
}

/** free the list with filenames
*/
void fastlist_free()
{
		for(unsigned int i= 0;i<memberof(playlist);++i)
          if(playlist[i].filename)
          {
            free(playlist[i].filename);
            playlist[i].filename = NULL;
          }
}

/* Search all files in a directory matching
with *.mp3 as file name.
@param const char * strDirName - the current directory
@param const char * filter - the search wildcard filter e.g. "*.mp3"
@param unsigned int& nFiles - Count founded files
@param unsigned int& nDir - current founded directories with files*/
bool fastlist_getfiles(const char * strDirName,const char * filter,unsigned int& nFiles,unsigned int nDir)
{
	int found;
	struct filefind ff;
	char szFind[MAXPATH];
	char szFile[MAXPATH];
	bool bRet = true;

   	//printf ("search on %s\n", strDirName);

    strncpy(szFind,strDirName,sizeof(szFind));
	strncat(szFind,filter,sizeof(szFind));

	found = fast_findfirst(szFind, &ff, 0);
	while (!found && nFiles<memberof(playlist) && bRet)
	{
		if (*ff.filename != '.' &&
			!(ff.fileattr & FA_DIREC))
		{
			strncpy(szFile,strDirName,sizeof(szFile));
			strncat(szFile,ff.filename,sizeof(szFile));
			strncat(szFile,".",sizeof(szFile));
			strncat(szFile,ff.fileext,sizeof(szFile));

			// duplicate founded filename in the playlist
            playlist[nFiles].filename = strdup(szFile);
            playlist[nFiles].nPlayed = 0;
            playlist[nFiles].nDirectory  = nDir;
			playlist[nFiles].nFilesize = ff.filesize;

//          printf("%3d. %s\n", nFiles, szFile);

			//Check successed duplicate String
			if(playlist[nFiles].filename)
			{
				if(0 == ++nFiles % 16) //Display only on 16th times
                {
					char szBuffer[80] = "Erzeuge Liste. ";
					itoa(nFiles,szBuffer + strlen(szBuffer),10);
	        	    strncat(szBuffer," Titel gefunden.",sizeof(szBuffer));
            		if(VK_POWER == display(szBuffer,nFiles,memberof(playlist),VK_INFO))
            	   	{
        	           	bRet = false;
    	               	break;
	               	}
                }
            }
            else	//No Memory ?
            {
				perror(szFile);
              	bRet = false;
            	break;
            }
		}
		found = fast_findnext(&ff);
	}
   fast_finddone(&ff);

   return bRet;
}

/** Search recursive all directories
this function call herself for recursive,
is a stackoverflow on SC12 possible ?
max. possible stackdeep is MAXDIR / 2 -> 33 if any use 1 char
@param const char * strDirName - the current directory
@param const char * filter - the search wildcard filter e.g. "*.mp3"
@param unsigned int& nFiles - Count founded files
@param unsigned int& nDir - Count founded directories with files (it's ignore empty dirs)*/
bool fastlist_getdirectorys(const char * strDirName,const char * filter,unsigned int& nFiles,unsigned int& nDir)
{
	//store foundet directories
	//i can't, one my SC12,
	//use more as one struct ffblk at same time :-(
	char *directorys[256] = {0};

	bool bRet = true;
	int found;
	unsigned int nDirs = 0;	//count directories
	struct filefind ff;   //findfirst...findnext struct

	char szDir[MAXPATH];
	char szFind[MAXPATH];

	// Remember the found file count
    unsigned int nBefore = nFiles;


	// init local current directorys
	for(unsigned int i= 0;i<memberof(directorys);++i)
			directorys[i] = NULL;

//	printf ("search on %s\n", strDirName);

	// Get first all files from this directory
	bRet = fastlist_getfiles (strDirName,filter,nFiles,nDir);

    //Count founded directories with files (it's ignore empty dirs)
	if(nBefore!=nFiles)
	    ++nDir;

	// Build search string "B:/PATH/PATH/*.*
	strncpy(szFind,strDirName,sizeof(szFind));
	strncat(szFind,"*.*",sizeof(szFind));

	// Find all directories
	found = fast_findfirst(szFind, &ff, FA_DIREC);
	while (!found && nFiles<memberof(playlist) && nDirs < memberof(directorys) && bRet)
	{
		if (*ff.filename != '.' &&
			(ff.fileattr & FA_DIREC))
		{
			// Limit Path to 66 charakters
   			if(MAXDIR - 1 > strlen(strDirName) + strlen(ff.filename))
			{

				// Build new search string "B:/PATH/PATH/FOUND/
				strncpy(szDir,strDirName,sizeof(szDir));
				strncat(szDir,ff.filename,sizeof(szDir));
				strncat(szDir,"\\",sizeof(szDir));

				directorys[nDirs] = strdup(szDir);	//Store directory name

				if(directorys[nDirs])  //check duplicate String
   	        	++nDirs;
		      else	//No Memory ?
         	{
					perror(szDir);
					bRet = false;
   	        	break;
	         }
         }
         else // Limit Path to 66 character
        	{
					perror(szDir);
					bRet = false;
   	        	break;
         }
		}
		// Search again
		found = fast_findnext(&ff);
	}
   fast_finddone(&ff);

	// Search on all directories for subdirectories and files
	for(unsigned int i = 0; i < nDirs && bRet && i < memberof(directorys) ;++i)
	{
		bRet = fastlist_getdirectorys (directorys[i],filter,nFiles,nDir);
	}
		// free used memory
	for(unsigned int i = 0; i < nDirs && i < memberof(directorys) ;++i)
	{
		free(directorys[i]);
        directorys[i] = NULL;
	}

   return bRet;
}

/** Search recursive all directories for all *.mp3
*  \param dir  Pointer to name of startpoint e.g "B:\PATH\"
*  \returns counts founded *.mp3 files.
*/
unsigned int fastlist_get(const char * szDirectory)
{
        char szFind[MAXPATH];
        strncpy(szFind,szDirectory,sizeof(szFind));

        // Add if missing backslash on end of path
        if(*(szFind + strlen(szFind) - 1) != '\\')
               strncat(szFind,"\\",sizeof(szFind));

        //counter for founded directories with files (it's ignore empty dirs)
	unsigned int nDir = 0;
    	//counter for founded files
	unsigned int nFiles = 0;
	// Alloc and  clean memory
	fastlist_init();
	// Search on all directories and subdirectories for *.mp3 files
	if(fastlist_getdirectorys(szFind,"*.MP3",nFiles,nDir))
		return nFiles;
   	else
   		return 0;
}

#ifndef FFTEST
unsigned int getnextfile(unsigned int nIndex,unsigned int nFiles,unsigned int nPlaymode,int nVK_Pressed)
{
	unsigned int nBegin = 0;
	unsigned int nEnd = nFiles;

	bool forward = (nVK_Pressed != VK_PREV);

	//Sanity check
   	if (nIndex >= nEnd)
	{
		if(forward)
			nIndex = nBegin;
		else
		 	nIndex = nEnd - 1;
	}

    unsigned int nDirectory = playlist[nIndex].nDirectory;
	if(nVK_Pressed == VK_PREVDIR )
	{
		// Search prev group
		// start from last
		unsigned int i = nIndex;
		for(;i<nFiles;--i)
     		if(playlist[i].nDirectory != nDirectory)
	           	break;
		// Warp around
	    if(i>=nFiles)
			for(i = nFiles - 1;i<nFiles;--i)
 	    		if(playlist[i].nDirectory != nDirectory)
	 	          	break;
	    // Found
	    if(i<nFiles)
	    {
			// Search first title in group
			for(;i>0;--i)
   		  		if(playlist[i-1].nDirectory != playlist[i].nDirectory)
	    	       	break;
	    		return i;
	    }
		// fallback to VK_PREV
	    forward = false;
	}
	else if(nVK_Pressed == VK_NEXTDIR)
	{
		// Search next group
		// start from last
		unsigned int i = nIndex;
		for(;i<nFiles;++i)
     		if(playlist[i].nDirectory != nDirectory)
	           	break;
		// Warp around
	    if(i>=nFiles)
			for(i = 0;i<nIndex;++i)
 	    		if(playlist[i].nDirectory != nDirectory)
	 	          	break;
	    // Found
	    if(i<nFiles)
	    	return i;
		// fallback to VK_NEXT
	    forward = true;
	}

		if(nPlaymode == RANDOMGROUP || nPlaymode == REPEATGROUP)
		{
			bool bFound = false;

			for(unsigned int i = 0;i<nFiles;++i)
			{
         	if(!bFound && playlist[i].nDirectory == nDirectory)
            {
            	nBegin = i;
                bFound = true;
            }
            if(bFound)
            {
                if(playlist[i].nDirectory != nDirectory)
                  break;
            	nEnd = i + 1;
            }
		 }
         if(!bFound) // Group not found is really
         {
			nBegin = 0;
			nEnd = nFiles;
         }
    }

    switch(nPlaymode)
    {
		case RANDOMGROUP:
        case RANDOM:
		{
			unsigned int nOld = nIndex;
            unsigned int nPlayed = playlist[nIndex].nPlayed;
        	// find a unplayed file what is not the next file
            for(unsigned int i = nBegin;i<nEnd;++i)
            {
                do
                {	// Generate random Nummer
   	               	nIndex = nBegin + rand() % (nEnd - nBegin);
                }	// check diff between new and old file bigger as nFiles/4
                while(abs(nOld - nIndex) < (nIndex >> 2));

                if(nPlayed > playlist[nIndex].nPlayed)
					break;
				// Found unplayed file
				}
				break;
			}

			case REPEAT:
			{
				break;
			}
        	case REPEATGROUP:
			case PLAY:
  	      	{
				if(forward)
				{
    	        	++nIndex;
					if (nIndex >= nEnd)
						nIndex = nBegin;
				}
				else
				{
	 		       	--nIndex;
					if (nIndex >= nEnd)
						nIndex = nEnd - 1;
				}
			break;
			}
		}//switch(nPlaymode)
	return nIndex;
}

unsigned int playfile(unsigned int iIndex,int nPlaymode)
{
	if(iIndex <  memberof(playlist)
      && playlist[iIndex].filename )
	{
		if(nPlaymode == RANDOM || nPlaymode == RANDOMGROUP) // Increment retry counter only on random mode
			++playlist[iIndex].nPlayed;

//		printf("%3d %s rty:%d grp:%d  \n",iIndex,playlist[iIndex].filename,playlist[iIndex].nPlayed,playlist[iIndex].nDirectory);
		return play(playlist[iIndex].filename,playlist[iIndex].nFilesize);
	}
	return VK_NEXT; //out off index ?
}

static void getfiletitle(FILE *f,char* title,unsigned int nMaxChar,const char* name)
{
	ID3v1Tag  tag;
	
	if (ID3v1_ReadTag(f, &tag))
	{
		// Build title "artist - title (comment)"

		if(strlen(tag.artist)) // Only if title present
			strncpy( title, tag.artist , nMaxChar);

		if(strlen(title)) // add if not empty
		{
			strncat( title, " - " , nMaxChar);
		}
		strncat( title, tag.title , nMaxChar);

		if(strlen(tag.comment))	// add comment if present
		{
			strncat( title, " (" , nMaxChar);
			strncat( title, tag.comment , nMaxChar);
			strncat( title, ")" , nMaxChar);
		}
	}

	// Can't read IDTag or empty IDTag, use nice Filename
	if(!strlen(title))
	{
		// conver filename from "b:\path\to\FILENAME.MP3
                // to title "path-to-FILENAME"
		const char * s = name;
		// Remove Drivename e.g. "b:\"
		for(;s && *s;++s)
		{
			if(*s == '\\' || *s == '/')
			{
				strncpy( title, s+1, nMaxChar );
				break;
			}
		}

		// Don't found drivename use pur filename
		if(*title == '\0')
			strncpy( title, name, nMaxChar );

		if(strlen(title))
		{
			//Remove Fileextension .mp3
			char* pe = title + strlen(title) - 1 ;
			for(;pe && *pe && pe != title ;--pe)
			{
				if(*pe == '.')
				{
					*pe = '\0';
					break;
				}
			}

			// Replace Slash with '-'
			char* pr = title;
			for(;pr && *pr;++pr)
			{
				if(*pr == '\\' || *pr == '/' )
				{
					*pr = '-';
				}
			}
		}
	}
}

unsigned int play(const char  * filename, unsigned long flen)
{
	//RC-return code
	unsigned int nVK_Pressed = VK_NEXT; //default play next file on leave

	FILE *f = fopen(filename, "rb");
	if (f)
	{
		char sz1stLine[80];
                *sz1stLine = '\0';
		
		getfiletitle(f,sz1stLine,sizeof(sz1stLine),filename);
		
		VS_Reset();
		BUF_Play();
		while (!(feof(f) && BUF_GetState() == BUF_STALLED))
		{
			// Read data from mp3 file
			//
			long  len, len1,  len2;
			char *buf1, *buf2;

			BUF_Lock( -1, min(16384L, BUF_GetFreeBytes()),
				&buf1, &len1, &buf2, &len2 );
			len = fread(buf1, 1, len1, f);
			if (buf2)
				len += fread(buf2, 1, len2, f);
			BUF_Unlock(len);

			long fpos =  ftell(f) - BUF_GetCount();

			// Update display and get remote
			int nRC = display(sz1stLine,fpos,flen);

			if(nRC == VK_POWER || nRC & VK_SKIPFUNCTION || nRC == VK_JUMP)
			{
				BUF_Clear();
				VS_Reset();
				nVK_Pressed = nRC;
				break;
			}

			if(nRC == VK_PLAY && !VS_IsPlaying() )
			{
				VS_Reset();
				BUF_Play();
			}
			if((nRC == VK_PAUSE || nRC == VK_STOP) && VS_IsPlaying())
			{
				BUF_Clear();
				VS_Reset();
				if(nRC == VK_STOP)
				{
					fseek(f,0, SEEK_SET);
				}
			}

			if(nRC == VK_FAST_REWIND)
			{
				BUF_Clear();
				BUF_Play();
				VS_Reset();

				long offset = (flen/20);
				if(fpos - offset > 0)
					fseek(f, -offset, SEEK_CUR);
				else
				{
					nVK_Pressed = VK_PREV;
					break;
				}
			}
			if(nRC == VK_FAST_FORWARD)
			{
				BUF_Clear();
				BUF_Play();
				VS_Reset();

				long offset = (flen/20);
				if((unsigned long)(fpos + offset) < flen)
					fseek(f, offset, SEEK_CUR);
				else
				{
					nVK_Pressed = VK_NEXT;
					break;
				}
			}
			// be nice to other apps
			//
			RTX_Sleep_Time(100);
		}
		fclose(f);
	}
	else
	{
		perror(filename);
	}

	return nVK_Pressed;
}
#endif //FFTEST
